/*
 * Framebuffer setup routines here.
 * Contact Jon Masters <jcm@jonmasters.org>.
 */

#include <asm/ppc/processor.h>
#include <const.h>
#include <hal/fbvideo/fbvideo.h>
#include <hal/fbvideo/fbvideo_table.h>
#include <hal/ofprom/prom.h>
#include <mm/mmu.h>

extern unsigned char vga_font[cmapsz];

/* extern void *prom_mmu; */

extern void *prom_finddevice(const char *device_name);
extern void *prom_open(const char *device_path);
extern int prom_instance_to_path(const char *device_name, char *device_path);
extern int prom_getprop(const char *device_name, const char *property_name,
			void *property_value, int property_length);
extern int prom_mmu_translate(void *prom_stdout, const char *mmu_handle,
			      const unsigned long virtual_address,
			      unsigned long *physical_address);
extern void prom_write(void *prom_stdout, const char *message);
extern void prom_write_hex(void *prom_stdout, unsigned int value);
extern void prom_exit(void);

extern char *strcpy(char *destination, const char *source);
extern char *strncpy(char *destination, const char *source, unsigned long size);
extern unsigned long strlen(const char *string);

extern void *memcpy(void *destination, const void *source, unsigned long size);

extern void fbvideo_write(const char *message);
extern void fbvideo_clear(unsigned long colour);

fbvideo_info_t fbvideo_table[FBVIDEO_DISPLAY_MAX];
fbvideo_info_t fbvideo_default;

void *fbvideo_device_instance;
void *fbvideo_device_package;

void *fbvideo_prom_stdout;

/* void fbvideo_BAT(void *prom_stdout); */

unsigned long fbvideo_BATL;
unsigned long fbvideo_BATU;

void fbvideo_scroll(void);

void fbvideo_init(void *prom_stdout, void *prom_mmu) {

  char prom_stdout_device_path[PROM_INSTANCE_TO_PATH_MAX_LENGTH];

  prom_instance_to_path(prom_stdout, prom_stdout_device_path);

  prom_write(prom_stdout,"fbvideo: ");
  prom_write(prom_stdout, prom_stdout_device_path);
  prom_write(prom_stdout, "\r\n");
  
  prom_write(prom_stdout, "Opening fbvideo...");
  fbvideo_device_instance = prom_open(prom_stdout_device_path);
  prom_write(prom_stdout, "Opened fbvideo\r\n");

  prom_write(prom_stdout, "Framebuffer device opened at: ");
  prom_write(prom_stdout, prom_stdout_device_path);
  prom_write(prom_stdout,"\r\n");

  prom_write(prom_stdout, "Calling finddevice...");
  fbvideo_device_package = (void *)prom_finddevice(prom_stdout_device_path);
  prom_write(prom_stdout, "done\r\n");

  prom_write(prom_stdout, "fbvideo_device_instance: ");
  prom_write_hex(prom_stdout, (unsigned long)fbvideo_device_instance);
  prom_write(prom_stdout, "\r\n");
  
  prom_write(prom_stdout, "fbvideo_device_package: ");
  prom_write_hex(prom_stdout, (unsigned long)fbvideo_device_package);
  prom_write(prom_stdout, "\r\n");

  prom_write(prom_stdout, "Setting fbvideo_table[0].name now...");
  strncpy(fbvideo_table[FBVIDEO_DISPLAY_ZERO].name,"UNINITIATED DISPLAY",20);
  prom_write(prom_stdout, "done\r\n");

  prom_getprop(fbvideo_device_package,"name",fbvideo_table[FBVIDEO_DISPLAY_ZERO].name,sizeof(fbvideo_table[FBVIDEO_DISPLAY_ZERO].name));
  prom_write(prom_stdout, "name: ");
  prom_write(prom_stdout, fbvideo_table[FBVIDEO_DISPLAY_ZERO].name);
  prom_write(prom_stdout, "\r\n");

  prom_getprop(fbvideo_device_package,"width",&fbvideo_table[FBVIDEO_DISPLAY_ZERO].width,sizeof(fbvideo_table[FBVIDEO_DISPLAY_ZERO].width));
  prom_write(prom_stdout, "width: ");
  prom_write_hex(prom_stdout, fbvideo_table[FBVIDEO_DISPLAY_ZERO].width);
  prom_write(prom_stdout, "\r\n");

  prom_getprop(fbvideo_device_package,"height",&fbvideo_table[FBVIDEO_DISPLAY_ZERO].height,sizeof(fbvideo_table[FBVIDEO_DISPLAY_ZERO].height));
  prom_write(prom_stdout, "height: ");
  prom_write_hex(prom_stdout, fbvideo_table[FBVIDEO_DISPLAY_ZERO].height);
  prom_write(prom_stdout, "\r\n");

  prom_getprop(fbvideo_device_package,"depth",&fbvideo_table[FBVIDEO_DISPLAY_ZERO].depth,sizeof(fbvideo_table[FBVIDEO_DISPLAY_ZERO].depth));
  prom_write(prom_stdout, "depth: ");
  prom_write_hex(prom_stdout, fbvideo_table[FBVIDEO_DISPLAY_ZERO].depth);
  prom_write(prom_stdout, "\r\n");

  prom_getprop(fbvideo_device_package,"linebytes",&fbvideo_table[FBVIDEO_DISPLAY_ZERO].linebytes,sizeof(fbvideo_table[FBVIDEO_DISPLAY_ZERO].linebytes));
  prom_write(prom_stdout, "linebytes: ");
  prom_write_hex(prom_stdout, fbvideo_table[FBVIDEO_DISPLAY_ZERO].linebytes);
  prom_write(prom_stdout, "\r\n");

  prom_getprop(fbvideo_device_package,"address",&fbvideo_table[FBVIDEO_DISPLAY_ZERO].address,sizeof(fbvideo_table[FBVIDEO_DISPLAY_ZERO].address));
  prom_write(prom_stdout, "address: ");
  prom_write_hex(prom_stdout, fbvideo_table[FBVIDEO_DISPLAY_ZERO].address);
  prom_write(prom_stdout, "\r\n");

  prom_mmu_translate(prom_stdout,prom_mmu,
		     (unsigned long)fbvideo_table[FBVIDEO_DISPLAY_ZERO].address,
		     &fbvideo_table[FBVIDEO_DISPLAY_ZERO].virtual);
  
  fbvideo_table[FBVIDEO_DISPLAY_ZERO].X = 0;
  fbvideo_table[FBVIDEO_DISPLAY_ZERO].Y = 0;
  
  fbvideo_table[FBVIDEO_DISPLAY_ZERO].X_limit = 
    fbvideo_table[FBVIDEO_DISPLAY_ZERO].width / 8;
  fbvideo_table[FBVIDEO_DISPLAY_ZERO].Y_limit =
    fbvideo_table[FBVIDEO_DISPLAY_ZERO].height / 16;

  fbvideo_table[FBVIDEO_DISPLAY_ZERO].background_colour = 0x00000000UL;
  fbvideo_table[FBVIDEO_DISPLAY_ZERO].foreground_colour = 0x0F0F0F0FUL;
  fbvideo_table[FBVIDEO_DISPLAY_ZERO].enabled = 1;

  prom_write(prom_stdout, "Initialised framebuffer: ");
  prom_write(prom_stdout, fbvideo_table[FBVIDEO_DISPLAY_ZERO].name);
  prom_write(prom_stdout, "\r\n");

  /* Eventually even I do get fed up with rediculously long symbol names. */

  prom_write(prom_stdout, "Setting default framebuffer device...");
  fbvideo_default = fbvideo_table[FBVIDEO_DISPLAY_ZERO];
  prom_write(prom_stdout, "done\r\n");

  prom_write(prom_stdout, "Framebuffer size: ");
  prom_write_hex(prom_stdout, fbvideo_table[FBVIDEO_DISPLAY_ZERO].X_limit);
  prom_write(prom_stdout, "x");
  prom_write_hex(prom_stdout, fbvideo_table[FBVIDEO_DISPLAY_ZERO].Y_limit);
  prom_write(prom_stdout, "\r\n");

  fbvideo_prom_stdout = prom_stdout;

  /* Calculate BAT entry now. */

  /* fbvideo_BAT(prom_stdout); */

}

unsigned long expand_bits[16] = {
  0x00000000,
  0x000000ff,
  0x0000ff00,
  0x0000ffff,
  0x00ff0000,
  0x00ff00ff,
  0x00ffff00,
  0x00ffffff,
  0xff000000,
  0xff0000ff,
  0xff00ff00,
  0xff00ffff,
  0xffff0000,
  0xffff00ff,
  0xffffff00,
  0xffffffff
};

void fbvideo_scroll(void) {

  unsigned long *fb_address_source                    = (unsigned long *)0;
  unsigned long *fb_address_destination               = (unsigned long *)0;
  int source_line, destination_line, current_word     = (int)0;

  for (source_line=16,destination_line=0;
       source_line<fbvideo_default.height;
       source_line++,destination_line++) {
    fb_address_source = (unsigned long *)(fbvideo_default.address +
					  (source_line *
					   fbvideo_default.linebytes));
    fb_address_destination = (unsigned long *)(fbvideo_default.address +
					       (destination_line *
						fbvideo_default.linebytes));
    for (current_word=0;
	 current_word<fbvideo_default.linebytes;
	 current_word++) {
      /* fb_address_source = (unsigned long *)(fbvideo_default.address +
       *				    (current_word +
       *				     (source_line *
       *				      fbvideo_default.linebytes)));
       * fb_address_destination = (unsigned long *)(fbvideo_default.address +
       *					 (current_word +
       *              				  (destination_line *
       *				       	   fbvideo_default.linebytes)));
       */
      /*fb_address_source+=current_word;
       *fb_address_destination+=current_word;
       *fb_address_destination[0] = fb_address_source[0];
       */
      /*fb_address_destination[1] = fb_address_source[1]; */

      *(fb_address_destination+current_word) = *(fb_address_source+current_word);

    }
  }
  fb_address_destination = (unsigned long *)(fbvideo_default.address +
					     ((fbvideo_default.height-16) *
					      fbvideo_default.linebytes));
  for (current_word=0;
       current_word<fbvideo_default.linebytes*16;
       current_word++,fb_address_destination++) {
    fb_address_destination[0] = fbvideo_default.background_colour;
  }
}

void fbvideo_write_char(unsigned long *address, const unsigned char c) {

  /* unsigned long fg              = 0x0F0F0F0FUL; */
  /* unsigned long bg              = 0x00000000UL; */
  int i,bits                    = 0;
  unsigned char *font           = &vga_font[((unsigned long)c) * 16];
  
  /* prom_write(fbvideo_prom_stdout,"Printing character: ");
   * prom_write_hex(fbvideo_prom_stdout,(unsigned long)c);
   * prom_write(fbvideo_prom_stdout,"\r\n");
   */

  switch(c) {
  case '\r':
    fbvideo_default.X = 0;
    break;
  case '\n':
    fbvideo_default.Y++;
    fbvideo_default.X = 0;
    break;
  default:

    for (i=0; i < 16; ++i) {
      /*
       * prom_write(fbvideo_prom_stdout,"Loop entry (");
       * prom_write_hex(fbvideo_prom_stdout,i);
       * prom_write(fbvideo_prom_stdout,")\r\n");
       */
      
      bits = *font++;
      address[0] = (expand_bits[bits >> 4] &
		    fbvideo_default.foreground_colour) ^ fbvideo_default.background_colour;
      address[1] = (expand_bits[bits & 0xf] & fbvideo_default.foreground_colour) ^ fbvideo_default.background_colour;
      address = (unsigned long *)((char *)address + fbvideo_default.linebytes);
      
      /* prom_write(fbvideo_prom_stdout,"Loop exit (");
       * prom_write_hex(fbvideo_prom_stdout,(unsigned long)address);
       * prom_write(fbvideo_prom_stdout,")\r\n");
       */
    }
    fbvideo_default.X++;
  }

  if (fbvideo_default.X >= fbvideo_default.X_limit) {
    fbvideo_default.X = 0;
    fbvideo_default.Y++;
  }

  if (fbvideo_default.Y >= fbvideo_default.Y_limit) {
    /* Scroll the framebuffer up. */
    /* fbvideo_default.Y = 0; */
    fbvideo_scroll();
    fbvideo_default.Y--; /* Stay on the bottom line of the screen. */
  }
}

void fbvideo_write(const char *message) {

  /* The font rendering code loop is based on btext.c in Linux since we
   * use the vga_font here.
   */

  unsigned long length          = 0;
  unsigned long printed         = 0;  
  char *message_ptr             = (char *)message;
  unsigned long *fb_address     = (unsigned long *)0;

  if (!fbvideo_default.enabled) {
    return;
  }

  length = strlen(message); /* Find out how much we have to write. */


  for (printed = 0; printed < length; printed++,message_ptr++) {

    /* Multiply width by eight to shift enough to fill character. */
    /* Multiply height by sixteen to shift enough to fill character. */

    fb_address = (unsigned long *)(fbvideo_default.address +
				   (fbvideo_default.X*8 +
				    (fbvideo_default.Y*16 *
				     fbvideo_default.linebytes)));
    /* prom_write(fbvideo_prom_stdout,"fb_address: ");
     * prom_write_hex(fbvideo_prom_stdout,(unsigned long)fb_address);
     * prom_write(fbvideo_prom_stdout,"\r\n");
     */

    /* prom_write(fbvideo_prom_stdout,"Call fbvideo_write_char\r\n"); */
    fbvideo_write_char(fb_address,*message_ptr);
    /* prom_write(fbvideo_prom_stdout,"Return fbvideo_write_char\r\n"); */

    /* prom_write(fbvideo_prom_stdout,"Position: ");
     * prom_write_hex(fbvideo_prom_stdout,fbvideo_default.X);
     * prom_write(fbvideo_prom_stdout,"x");
     * prom_write_hex(fbvideo_prom_stdout,fbvideo_default.Y);
     * prom_write(fbvideo_prom_stdout,"\r\n");
     */

  }
}

void fbvideo_write_hex(unsigned int value) {
  
  char buf[16];
  int i, current_char;

  for (i = 0; i < 8; i++) {
    current_char  = (value >> ((7-i)*4)) & 0xf;
    current_char += (current_char >= 10) ? ('a' - 10) : '0';
    buf[i] = current_char;
  }
  buf[i] = 0;

  fbvideo_write(buf);
}

void fbvideo_clear(unsigned long colour) {

  unsigned long *fb_address = (unsigned long *)0;
  /*  unsigned long bg          = 0x00000000UL; */
  int i,j                   = 0;

  /* Clear the framebuffer of evil now. */

  for (i=0; i<fbvideo_default.height; i++) {
    for (j=0; j<fbvideo_default.width; j++) {
      fb_address = (unsigned long *)(fbvideo_default.address +
				     (j + (i * fbvideo_default.linebytes)));
      fb_address[0] = (unsigned long)colour;
    }
  }

  fbvideo_default.X = 0;
  fbvideo_default.Y = 0;

}
