/*
 * Open Firmware Client Interface handling routines. These routines build a
 * prom_args structure and then call in to the PROM using the client_interface.
 *
 * Copyright (C) 2003 Jon Masters.
 *
 * This code is relatively simple at this stage and could do with a cleanup
 * along the lines of that used in yaboot, Linux or even NetBSD (weird stuff).
 *
 * Need to typedef an ihandle and phandle though for now let us overuse long
 * and void data types because that seems like the nutty thing to do.
 *
 * Other PowerPC operating systems do not seem to preserve the OpenFirmware
 * environment however it is possible to do so with careful mmu setup.
 *
 * prom_finddevice("/devine-inspiration");
 */

#include <hal/ppc/processor.h>
#include <hal/ofprom/prom.h>

#define FLUX_VERSION "0.0.1"

/* Apple engineers scare me with their clean code. They use funky unions and
 * descriptive variables though we do not bother with that here yet.
 */

struct prom_args {
  const char *service;
  int number_args;
  int number_args_returned;
  void *args[10];
};

unsigned long prom_setup(int r3, int r4, int prom_client);
void *prom_finddevice(const char *device_name);
int prom_getprop(const char *device_name, const char *property_name,
		 void *property_value, int property_length);
int prom_instance_to_path(const char *device_name, char *path_name);
void *prom_open(const char *device_path);
int prom_mmu_translate(void *prom_stdout, const char *mmu_handle,
		       const unsigned long virtual_address,
		       unsigned long *physical_address);
/* void prom_write(void *prom_stdout, const char *message, int length); */
void prom_write(void *prom_stdout, const char *message);
void prom_write_hex(void *prom_stdout, unsigned int value);
void prom_dispinfo(void *prom_stdout);
void prom_exit(void);
void prom_quiesce(void);
int prom_milliseconds(void);

void (*client_interface)(struct prom_args *);

extern char _start,_end;

extern unsigned long reloc_offset();
extern void reloc_got2();

extern void fbvideo_init(void *prom_stdout, void *prom_mmu);
extern void fbvideo_write(const char *message);

unsigned long physical_start_address;
unsigned long physical_end_address;

unsigned long prom_setup(int r3, int r4, int prom_client) {

  void *prom_chosen;
  void *prom_stdin;
  void *prom_stdout;
  /*  void *prom_bootpath; */
  /*  void *prom_bootargs; */
  void *prom_memory;
  void *prom_mmu;

  /* unsigned long offset = reloc_offset(); */

  /* void (*client_interface)(struct prom_args *); */

  /* unsigned long offset = reloc_offset(); */
  
  /* Open Firmware client interface initialisation code.
   *
   * The argument list items "r3" and "r4" are not used, Apple call these
   * "unused", Linux uses register names, which seems a little more logical.
   *
   */

  /* doors:
     
     goto doors;
     
     if (offset != 0) {
     goto doors;
     } */
  
  /* The magical pandas will not get us, the problem is in relocation. */
  
  /* reloc_got2(offset); Remember to undo this later on! */

  /* remember this is the wrong client_interface! */
  client_interface = (void (*)(struct prom_args *))prom_client;
  
  if ((void *)client_interface <= (void *)0) {
    /* We are completely fucked. We cannot even exit cleanly now! */
  moon:
    goto moon; /* in a spaceship please...round and round forever... */
  }
  
  /* doors:
     goto doors; */
  
  prom_chosen = (void *)prom_finddevice("/chosen");
  
  /* doors1:
     goto doors1; */

  prom_getprop(prom_chosen,"stdout",&prom_stdout,sizeof(prom_stdout));
  
  /* doors2:
     goto doors2; */
  
  prom_write(prom_stdout,"OpenFirmware Startup\r\n");
  prom_write(prom_stdout,"--------------------\r\n");
  prom_write(prom_stdout,"Software Release: ");
  prom_write(prom_stdout,FLUX_VERSION);
  prom_write(prom_stdout,"\r\n");

  prom_dispinfo(prom_stdout);

  prom_write(prom_stdout,"Reading required OpenFirmware devices...");
  prom_getprop(prom_chosen,"stdin", &prom_stdin,sizeof(prom_stdin));
  prom_getprop(prom_chosen,"memory", &prom_memory,sizeof(prom_memory));
  prom_write(prom_stdout,"memory...");
  prom_getprop(prom_chosen,"mmu", &prom_mmu,sizeof(prom_mmu));
  prom_write(prom_stdout,"mmu");
  prom_write(prom_stdout,"...done\r\n");

  /*  prom_write(prom_stdout,"translate client_interface (");
      prom_write_hex(prom_stdout,(unsigned long)client_interface);
      prom_write(prom_stdout,"): ");
      prom_mmu_translate(prom_stdout,prom_mmu,
      (unsigned long)client_interface,&translation_address);
      prom_write_hex(prom_stdout,translation_address);
      prom_write(prom_stdout,"\r\n"); */

  prom_write(prom_stdout,"Running at virtual address:  ");
  prom_write_hex(prom_stdout,(unsigned long)&_start);
  prom_write(prom_stdout," - ");
  prom_write_hex(prom_stdout,(unsigned long)&_end);
  prom_write(prom_stdout,"\r\n");
  prom_write(prom_stdout,"Running at physical address: ");
  prom_mmu_translate(prom_stdout,prom_mmu,
		     (unsigned long)&_start,&physical_start_address);
  prom_write_hex(prom_stdout,physical_start_address);
  prom_write(prom_stdout," - ");
  prom_mmu_translate(prom_stdout,prom_mmu,
		     (unsigned long)&_end,&physical_end_address);
  prom_write_hex(prom_stdout,physical_end_address);
  prom_write(prom_stdout,"\r\n");

  if (((unsigned long)&_start == physical_start_address) &&
      ((unsigned long)&_end == physical_end_address)) {
    prom_write(prom_stdout,"Running with 1:1 mmu memory mapping.\r\n");
  } else {
    prom_write(prom_stdout,"Non-1:1 mmu memory mapping not supported.\r\n");
    prom_exit();
  }

  /* debug:
   * goto debug;
   */

  prom_write(prom_stdout,"Framebuffer information...");

  /* START Framebuffer initialisation. */

  fbvideo_init(prom_stdout,prom_mmu); /* Callout and use callbacks. */

  /* END Framebuffer initialisation. */

  prom_write(prom_stdout,"Let's do it now...\r\n");
  prom_write(prom_stdout,"Nuking OpenFirmware...");
  prom_quiesce();
  prom_write(prom_stdout,"done\r\n");

  /*  prom_write(prom_stdout,"exit");
   *  prom_exit();
   */

  return 0;

}

void *prom_finddevice(const char *device_name) {

  /* Return an ihandle pointer to the device specified in the device_name
   * path.
   */

  struct prom_args prom_args;

  prom_args.service = "finddevice";
  prom_args.number_args = PROM_FINDDEVICE_NUMBER_ARGS;
  prom_args.number_args_returned = PROM_FINDDEVICE_NUMBER_ARGS_RETURNED;
  prom_args.args[0] = (void *)device_name;
  prom_args.args[1] = 0; /* zero element */

  client_interface(&prom_args);
  return prom_args.args[1];

}

int prom_getprop(const char *device_name, const char *property_name,
		 void *property_value, int property_length) {

  /* Return the value of a property of a specified device_name.
   */

  struct prom_args prom_args;

  prom_args.service = "getprop";
  prom_args.number_args = PROM_GETPROP_NUMBER_ARGS;
  prom_args.number_args_returned = PROM_GETPROP_NUMBER_ARGS_RETURNED;
  prom_args.args[0] = (void *)device_name;
  prom_args.args[1] = (void *)property_name;
  prom_args.args[2] = (void *)property_value;
  prom_args.args[3] = (void *)property_length;
  prom_args.args[4] = 0; /* zero element */

  client_interface(&prom_args);
  return (int)prom_args.args[PROM_GETPROP_NUMBER_ARGS];

}

void *prom_open(const char *device_path) {

  /* Open a specified device_path.
   */

  struct prom_args prom_args;

  prom_args.service = "open";
  prom_args.number_args = PROM_OPEN_NUMBER_ARGS;
  prom_args.number_args_returned = PROM_OPEN_NUMBER_ARGS_RETURNED;
  prom_args.args[0] = (void *)device_path;
  prom_args.args[1] = 0; /* zero element */

  client_interface(&prom_args);
  return (void *)prom_args.args[PROM_OPEN_NUMBER_ARGS];

}

int prom_instance_to_path(const char *device_name, char *path_name) {

  /* Return the path of a specified device_name.
   */

  struct prom_args prom_args;

  prom_args.service = "instance-to-path";
  prom_args.number_args = PROM_INSTANCE_TO_PATH_NUMBER_ARGS;
  prom_args.number_args_returned = PROM_INSTANCE_TO_PATH_NUMBER_ARGS_RETURNED;
  prom_args.args[0] = (void *)device_name;
  prom_args.args[1] = (void *)path_name;
  prom_args.args[2] = (void *)PROM_INSTANCE_TO_PATH_MAX_LENGTH;
  prom_args.args[3] = 0; /* zero element */

  client_interface(&prom_args);
  return (int)prom_args.args[PROM_INSTANCE_TO_PATH_NUMBER_ARGS];

}

int prom_mmu_translate(void *prom_stdout, const char *mmu_handle,
		       const unsigned long virtual_address,
		       unsigned long *physical_address) {

  /* physical_address will contain the translation for virtual_address. */

  /* dfb will hate this one... yes it is ugly and not as clean as say the
   * Apple Darwin. Still it works now... */

  struct prom_args prom_args;

  prom_args.service = "call-method";
  prom_args.number_args = PROM_MMU_TRANSLATE_NUMBER_ARGS;
  prom_args.number_args_returned = PROM_MMU_TRANSLATE_NUMBER_ARGS_RETURNED;
  /* prom_args.args[0] = (void *)translation_array; */
  prom_args.args[0] = (void *)"translate";
  prom_args.args[1] = (void *)mmu_handle;
  prom_args.args[2] = (void *)virtual_address;
  prom_args.args[3] = (void *)1;
  prom_args.args[4] = 0; /* zero element */
  prom_args.args[5] = 0; /* zero element */
  prom_args.args[6] = 0; /* zero element */
  prom_args.args[7] = 0; /* zero element */

  client_interface(&prom_args);
  
  *physical_address = (unsigned long)prom_args.args[7];

  return (int)prom_args.args[PROM_MMU_TRANSLATE_NUMBER_ARGS];

}

/* void prom_write(void *prom_stdout, const char *message, int length) { */
void prom_write(void *prom_stdout, const char *message) {

  struct prom_args prom_args;

  int message_size = 0;
  const char *message_pointer = 0;

  prom_args.service = "write";
  prom_args.number_args = PROM_WRITE_NUMBER_ARGS;
  prom_args.number_args_returned = PROM_WRITE_NUMBER_ARGS_RETURNED;
  prom_args.args[0] = (void *)prom_stdout;
  prom_args.args[1] = (void *)message;
  prom_args.args[2] = (void *)0;
  prom_args.args[3] = 0; /* zero element */

  if ((void *)prom_stdout == 0) {
    return; /* There is no standard output instance handle here! */
  }

  for (message_pointer = message; *message_pointer != 0; message_pointer++) {
    message_size++;
  }

  prom_args.args[2] = (void *)message_size;

  client_interface(&prom_args);

}

void prom_write_hex(void *prom_stdout, unsigned int value) {

  struct prom_args prom_args;

  char buf[16];
  int i, current_char;

  /* prom_write(prom_stdout,"Entering Loop\r\n"); */

  for (i = 0; i < 8; i++) {
    /*    prom_write(prom_stdout,"Loop Entry\r\n"); */
    current_char  = (value >> ((7-i)*4)) & 0xf;
    current_char += (current_char >= 10) ? ('a' - 10) : '0';
    buf[i] = current_char;
  }

  /* prom_write(prom_stdout,"Leaving Loop\r\n"); */

  buf[i] = 0;

  prom_args.service = "write";
  prom_args.number_args = PROM_WRITE_NUMBER_ARGS;
  prom_args.number_args_returned = PROM_WRITE_NUMBER_ARGS_RETURNED;
  prom_args.args[0] = (void *)prom_stdout;
  prom_args.args[1] = (void *)buf;
  prom_args.args[2] = (void *)i;
  prom_args.args[3] = 0; /* zero element */

  if ((void *)prom_stdout == 0) {
    return; /* There is no standard output instance handle here! */
  }

  client_interface(&prom_args);

}

void prom_dispinfo(void *prom_stdout) {

  /* This will display some handy information to aid in debugging, such as the
   * current MSR - Machine Status - register value, the Processor Version
   * number and the status of memory management.
   */

  unsigned int msr,pvr,sdr1;

  unsigned int ibat0l, ibat0u, ibat1l, ibat1u, ibat2l, ibat2u, ibat3l, ibat3u;
  unsigned int dbat0l, dbat0u, dbat1l, dbat1u, dbat2l, dbat2u, dbat3l, dbat3u;

  prom_write(prom_stdout,"--- System Information ---\r\n");
  prom_write(prom_stdout,"\r\n");

  __asm__ __volatile__ ("mfmsr %0"          : "=r" (msr)    );
  __asm__ __volatile__ ("mfpvr %0"          : "=r" (pvr)    );
  __asm__ __volatile__ ("mfsdr1 %0"         : "=r" (sdr1)   );

  __asm__ __volatile__ ("mfibatl %0,0"      : "=r" (ibat0l) );
  __asm__ __volatile__ ("mfibatu %0,0"      : "=r" (ibat0u) );
  __asm__ __volatile__ ("mfibatl %0,1"      : "=r" (ibat1l) );
  __asm__ __volatile__ ("mfibatu %0,1"      : "=r" (ibat1u) );
  __asm__ __volatile__ ("mfibatl %0,2"      : "=r" (ibat2l) );
  __asm__ __volatile__ ("mfibatu %0,2"      : "=r" (ibat2u) );
  __asm__ __volatile__ ("mfibatl %0,3"      : "=r" (ibat3l) );
  __asm__ __volatile__ ("mfibatu %0,3"      : "=r" (ibat3u) );
  __asm__ __volatile__ ("mfibatl %0,0"      : "=r" (dbat0l) );
  __asm__ __volatile__ ("mfibatu %0,0"      : "=r" (dbat0u) );
  __asm__ __volatile__ ("mfibatl %0,1"      : "=r" (dbat1l) );
  __asm__ __volatile__ ("mfibatu %0,1"      : "=r" (dbat1u) );
  __asm__ __volatile__ ("mfibatl %0,2"      : "=r" (dbat2l) );
  __asm__ __volatile__ ("mfibatu %0,2"      : "=r" (dbat2u) );
  __asm__ __volatile__ ("mfibatl %0,3"      : "=r" (dbat3l) );
  __asm__ __volatile__ ("mfibatu %0,3"      : "=r" (dbat3u) );

  prom_write(prom_stdout,"   MSR: ");
  prom_write_hex(prom_stdout,msr);
  if (msr&(MSR_IR|MSR_DR)) {
    prom_write(prom_stdout," (IR) (DR)");
  }
  if (msr&(MSR_ME)) {
    prom_write(prom_stdout," (ME)");
  }
  if (msr&(MSR_FP)) {
    prom_write(prom_stdout," (FP)");
  }
  prom_write(prom_stdout,"\r\n");
  prom_write(prom_stdout,"   PVR: ");
  if (pvr&(PROCESSOR_VERSION_7455)) {
    prom_write(prom_stdout,"Motorola 7455");
  } else {
    prom_write_hex(prom_stdout,pvr);
  }
  prom_write(prom_stdout,"\r\n");
  prom_write(prom_stdout,"  SDR1: ");
  prom_write_hex(prom_stdout,sdr1);
  prom_write(prom_stdout,"\r\n");
  prom_write(prom_stdout,"IBAT0L: ");
  prom_write_hex(prom_stdout,ibat0l);
  prom_write(prom_stdout,"\r\n");
  prom_write(prom_stdout,"IBAT0U: ");
  prom_write_hex(prom_stdout,ibat0u);
  prom_write(prom_stdout,"\r\n");
  prom_write(prom_stdout,"IBAT1L: ");
  prom_write_hex(prom_stdout,ibat1l);
  prom_write(prom_stdout,"\r\n");
  prom_write(prom_stdout,"IBAT1U: ");
  prom_write_hex(prom_stdout,ibat1u);
  prom_write(prom_stdout,"\r\n");
  prom_write(prom_stdout,"IBAT2L: ");
  prom_write_hex(prom_stdout,ibat2l);
  prom_write(prom_stdout,"\r\n");
  prom_write(prom_stdout,"IBAT2U: ");
  prom_write_hex(prom_stdout,ibat2u);
  prom_write(prom_stdout,"\r\n");
  prom_write(prom_stdout,"IBAT3L: ");
  prom_write_hex(prom_stdout,ibat3l);
  prom_write(prom_stdout,"\r\n");
  prom_write(prom_stdout,"IBAT3U: ");
  prom_write_hex(prom_stdout,ibat3u);
  prom_write(prom_stdout,"\r\n");
  prom_write(prom_stdout,"DBAT0L: ");
  prom_write_hex(prom_stdout,dbat0l);
  prom_write(prom_stdout,"\r\n");
  prom_write(prom_stdout,"DBAT0U: ");
  prom_write_hex(prom_stdout,dbat0u);
  prom_write(prom_stdout,"\r\n");
  prom_write(prom_stdout,"DBAT1L: ");
  prom_write_hex(prom_stdout,dbat1l);
  prom_write(prom_stdout,"\r\n");
  prom_write(prom_stdout,"DBAT1U: ");
  prom_write_hex(prom_stdout,dbat1u);
  prom_write(prom_stdout,"\r\n");
  prom_write(prom_stdout,"DBAT2L: ");
  prom_write_hex(prom_stdout,dbat2l);
  prom_write(prom_stdout,"\r\n");
  prom_write(prom_stdout,"DBAT2U: ");
  prom_write_hex(prom_stdout,dbat2u);
  prom_write(prom_stdout,"\r\n");
  prom_write(prom_stdout,"DBAT3L: ");
  prom_write_hex(prom_stdout,dbat3l);
  prom_write(prom_stdout,"\r\n");
  prom_write(prom_stdout,"DBAT3U: ");
  prom_write_hex(prom_stdout,dbat3u);
  prom_write(prom_stdout,"\r\n");

  prom_write(prom_stdout,"/chosen/mmu regions");
  prom_write(prom_stdout,"\r\n");
  prom_write(prom_stdout,"available:\r\n");
  prom_write(prom_stdout,"\r\n");
  prom_write(prom_stdout,"existing:\r\n");
  prom_write(prom_stdout,"\r\n");
  prom_write(prom_stdout,"/chosen/memory regions");
  prom_write(prom_stdout,"\r\n");
  prom_write(prom_stdout,"available:\r\n");
  prom_write(prom_stdout,"\r\n");
  prom_write(prom_stdout,"reg:\r\n");
  prom_write(prom_stdout,"\r\n");

  prom_write(prom_stdout,"\r\n");
  prom_write(prom_stdout,"--- End Information ---\r\n");

}

void prom_exit(void) {

  struct prom_args prom_args;

  prom_args.service = "exit";
  prom_args.number_args = 0;
  prom_args.number_args_returned = 0;

  client_interface(&prom_args); /* never returns */

}

void prom_quiesce(void) {

  struct prom_args prom_args;

  prom_args.service = "quiesce";
  prom_args.number_args = 0;
  prom_args.number_args_returned = 0;

  client_interface(&prom_args); /* OpenFirmware will close open devices now */

}

int prom_milliseconds(void) {

  struct prom_args prom_args;

  prom_args.service = "milliseconds";
  prom_args.number_args = 0;
  prom_args.number_args_returned = 1;

  client_interface(&prom_args); /* Return Milliseconds. */

  return (int)prom_args.args[0];

}
