/*
 * Copyright (C) 2003 Jon Masters <jcm@jonmasters.org>
 *
 * Low level startup code for PowerMac booting from Open Firmware. This code
 * is of alpha quality and has not been tested. Nobody would want to use this.
 *
 */

#include <const.h>
#include <hal/ppc/processor.h>
#include <asm/ppc/processor.h>
#include <hal/fbvideo/fbvideo_emergency.h>
#include <mm/mmu.h>


/* Linux PPC and indeed most other PowerPC Operating Systems are started by
 * a boot loader in most cases however let us keep things simple for now.
 *
 * Actually the above no longer applies since we need a bootloader to put us
 * somewhere before we relocate to our new home in physical memory.
 */

	.text
	.globl _start

_start:

	/* Open Firmware loads us at the load address and we enter here.
	 * At this point virtual->physical mapping is 1:1 and Open Firmware is
	 * responsible for MMU hashtable and assorted other system functions.
	 * r5 contains the client interface pointer for making prom calls.
	 */

	/* The order for initialisation is:
	 *  1). Find out what we are doing here
	 *  2). Check /options
	 *  3). Check /chosen
	 *  4). Setup stdout
	 *  4). Build a memory map
	 *  5). Claim memory (physical and logical claim)
	 *  6). Map memory in to one large contiguous chunk
	 *  7). Display hacks for framebuffer use later
	 *  8). Setup MSR (Interrupts off, Translation off)
	 *  9). Instantiate Exception Vectors
	 * 10). Move kernel to link address now (if required)
	 */
	
	/* bl prom_setup - setup open firmware client interface now */

/* This is head for debug purposes. It was generally useful to immediately be
 * able to branch to various locations though it is not necessary now. I cannot
 * be bothered to sort this out either. :-).
 */

	b genstart

/* The fun starts here. Do not fear the Giant Panda and he will not fear you.
 */

genstart:
	/* MMU is still on. Stupid thing. */
	/* bl ppc_mmu_disable */ /* TEST CODE. */
	/* bl ppc_mmu_enable */ /* TEST CODE. */
	bl setup_pmac					/* Call setup. */
	mr r24,r3                                       /* Preserve Address. */
	/* bl emergency_framebuffer_fill */
	/* bl ppc_mmu_disable */
	/* bl ppc_mmu_enable */
	/* bl emergency_framebuffer_fill */

	/* At this point the basic setup is completed and we are ready to
	 * relocate the kernel now. r3 contains the current physical address.
	 * and is used to calculate positions of data items below.
	 */

	/* The current I address is about to be invalidated as part of the
	 * MMU setup. This existence has been short though hopefully noble.
	 * This line was infected with a Microsoft Windows script worm.
	 */
	
	bl ppc_mmu_setup                                /* Setup Memory. */
	/* bl ppc_mmu_disable */
	/* bl ppc_mmu_enable */
	/* bl ppc_cache_setup */                        /* Cache Setup. */
	/* bl start_kernel */                           /* Start Kernel. */

/* We should never reach this point.
 * If we are here then my point about Giant Pandas taking over the
 * world has been well proven correct by now. Run as fast as you can.
 */

1:
	b 1b

/* Flush the TLBs by iterating through one page frame at a time. */
	
	.globl ppc_mmu_flush_tlbs
ppc_mmu_flush_tlbs:
	lis    r20, 0x40         /* Going to clear 256MB. */
1:
	addic. r20, r20, -0x1000 /* One page at a time please. */
	tlbie  r20
	blt    1b
	sync
	blr

/* PowerPC MMU Setup Routine. */

ppc_mmu_setup:

/* Disable the MMU first. */

	mfmsr r5
	ori   r0,0,MSR_DR|MSR_IR
	andc  r5,r5,0
	/* mtmsr r5 */ /* COMMENTED OUT FOR TESTING. */
	isync
/* Relocate the kernel code. */

	/* lis   r24,physical_start_address@ha */     /* start symbol. */
	/* ori   r24,r24,physical_start_address@l */
	lis   r25,_end@ha                       /* End of kernel image. */
	ori   r25,r25,_end@l

	lis   r26,SYSTEM_ZERO@ha
	ori   r26,r26,SYSTEM_ZERO@l

	subf  r27,r24,25
	mtctr r27
	
	/* Now that the required data has been loaded, call reloc_got2 to
	 * undo the previous got2 relocation as it is no longer required and
	 * must be changed to use the correct addresses now.
	 *
	 * THERE IS NO ACCESS TO GLOBAL DATA BEYOND THIS POINT.
	 * ANY REQUIRED VALUES FOR RELOCATION MUST BE LOADED ABOVE.
	 */

	/* bl    reloc_offset */
	/* neg   r3,r3 */                               /* Negative offset. */
	/* bl    reloc_got2 */

	/* DO NOT BRANCH AFTER THIS POINT. */

1:
	lwz   r0,0(r24)
	stw   r0,0(r26)
	addi  r24,r24,1
	addi  r26,r26,1
	bdnz  1b
	
	/* bl    prom_exit */
	
	lis   r0,1f@h
	ori   r0,r0,1f@l
	mtctr r0
	bctr
1:
	/* DEBUG: Relocate to previous run address and paint framebuffer. */
	
	/* bl    reloc_offset */
	/* bl    reloc_got2 */
	/* bl    emergency_framebuffer_fill */

/* Relocation over now cleanup. */

/* 1: */
/*	b 1b */
/* Enable the MMU last. */
	
	mfmsr r5
	ori   r0,r5,MSR_DR|MSR_IR
	mtmsr r5
	isync

	/* bl prom_exit */

	bl start_kernel

/* Relocate Kernel. */

ppc_relocate_kernel:

/* Turn off the MMU. */

ppc_mmu_disable:
	li    r0,0                  /* Clear register 0 first. */
	mfmsr r5
	ori   r0,0,MSR_DR|MSR_IR    /* Load the required bitmask. */
	andc  r5,r5,r0		    /* Complement and with MSR. */
	mtmsr r5                    /* Store the new MSR. */
	isync                       /* CSI Required. */
	blr

/* Turn on the MMU. */

ppc_mmu_enable:
	mfmsr r5
	ori   r0,r5,MSR_DR|MSR_IR
	mtmsr r5
	isync
	blr

/* An emergency routine similar in function to an LED attached to a serial
 * port. The framebuffer can be painted extremely quickly through this hard
 * coded assembly which works only on systems with the framebuffer base.
 */

emergency_framebuffer_fill:
	lis   r6,FBVIDEO_EMERG_LINECOUNT@ha
	addi  r6,r6,FBVIDEO_EMERG_LINECOUNT@l
	lis   r5,FBVIDEO_EMERG_LINEBYTES@ha /* Load number of bytes per line. */
	addi  r5,r5,FBVIDEO_EMERG_LINEBYTES@l
	mullw r5,r5,r6
	mtctr r5                            /* Setup the count register. */
	lis   r4,FBVIDEO_EMERG_BASE@ha      /* Load emergency fb. address. */
	addi   r4,r4,FBVIDEO_EMERG_BASE@l
	lis   r0,FBVIDEO_EMERG_COLOUR@ha    /* Load emergency fb. colour. */
	addi   r0,r0,FBVIDEO_EMERG_COLOUR@l
/*	mr    r3,r4
 *	mflr  30
 *	bl    fbvideo_write_hex
 *	mtlr  30
 *	blr
 */

loop:
	/* stb   r0,0(r4) */
	stw   r0,0(r4)
	addi  r4,r4,1
	bdnz  loop
	blr

	/* MAGIC MARKER.
	 * In the vein hope that one day I will have a hardware debugger or a
	 * simulator capable of running this code.
	 */
1:
	b 1b /* we should never reach here normally */

#if 0

exception_vectors:

/**************************** EXCEPTION VECTORS. *****************************/

	. = 0x100
	b	__exception_system_reset	/* Should not happen normally. */

	. = 0x200
	b	__exception_machine_check	/* Should not happen normally. */

	. = 0x300
	b	__exception_data_access			/* DSI probably alignment issue. */

	. = 0x400
	b	__exception_instruction_access		/* ISI probably alignment issue. */

	. = 0x500
	b	__exception_external			/* OpenPIC interrupt. */

	. = 0x600
	b	__exception_alignment			/* Alignment exception. */

	. = 0x700
	b	__exception_program_check		/* Program check. */

	. = 0x800
	b	__exception_floating_point		/* Floating Point Unavailable exception. */

	. = 0x900
	b	__exception_timer			/* Timer interrupt. */

	. = 0xc00
	b	__ppc_message_handler			/* User space sending message. */
	
	/* Skip software TLB assist exceptions here. */

	/* Skip unrequired exceptions here. */

	/* The limit of the exception space should be marked so that the
	 * kernel text continues after the exception space. */

	. = 0x3000

/********************************* HANDLERS ********************************/

	.globl __exception_system_reset
	.globl __exception_machine_check
	.globl __exception_data_access
	.globl __exception_instruction_access
	.globl __exception_external
	.globl __exception_alignment
	.globl __exception_program_check
	.globl __exception_floating_point
	.globl __exception_timer
	.globl __ppc_message_handler

/* SYSTEM RESET EXCEPTION NOT NORMALLY INVOKED ONCE RUNNING.
 * In the same way that Linux attempts to store useful debugging information
 * directly to memory we also write the current values of SRRs to RAM and
 * then enter and infinite loop waiting for user intervention. */

__exception_system_reset:
	li	r3,0
	mfspr	r0,SRR0
	stw	r0,0(r3)
	mfspr	r0,SRR1
	stw	r0,4(r3)
1:
	b	1b

__exception_machine_check:
	li	r3,0
	mfspr	r0,SRR0
	stw	r0,0(r3)
	mfspr	r0,SRR1
	stw	r0,4(r3)
1:
	b	1b

__exception_data_access:
	/* We do not handle this exception yet. */
1:
	b	1b

__exception_instruction_access:
	/* We do not handle this exception yet. */
1:
	b	1b
	
__exception_external:
	/* OpenPIC exception handling code. */
	/* bl openpic_exception */
	sync
	isync
	rfi	
__exception_alignment:
	/* Generic alignment exception. Not implemented yet and as a result
	 * the code has copious use of convieniently sized data types. */
1:
	b	1b

__exception_program_check:
	/* What actually causes this exception to occur anyway? */
1:	
	b	1b	
__exception_floating_point:
	/* We do not support floating point. */
1:
	b	1b

__exception_timer:
	/* A timer interrupt occured. */
	/* bl timer_exception */
	sync
	isync
	rfi
	
__ppc_message_handler:

/******************************** END ***************************************/

1:
	b	1b

#endif
