/*
 * Copyright (C) 2003 Jon Masters <jcm@jonmasters.org>
 *
 */

#include <hal/ppc/processor.h>

void sysinfo_display(void);

extern void fbvideo_write(const char *message);
extern void fbvideo_write_hex(unsigned int value);

void sysinfo_display(void) {
  
  unsigned int msr,pvr,sdr1;
  
  unsigned int ibat0l, ibat0u, ibat1l, ibat1u, ibat2l, ibat2u, ibat3l, ibat3u;
  unsigned int dbat0l, dbat0u, dbat1l, dbat1u, dbat2l, dbat2u, dbat3l, dbat3u;

  __asm__ __volatile__ ("mfmsr %0"          : "=r" (msr)    );
  __asm__ __volatile__ ("mfpvr %0"          : "=r" (pvr)    );
  __asm__ __volatile__ ("mfsdr1 %0"         : "=r" (sdr1)   );
  
  __asm__ __volatile__ ("mfspr %0,0x211"    : "=r" (ibat0l) );
  __asm__ __volatile__ ("mfspr %0,0x210"    : "=r" (ibat0u) );
  __asm__ __volatile__ ("mfspr %0,0x213"    : "=r" (ibat1l) );
  __asm__ __volatile__ ("mfspr %0,0x212"    : "=r" (ibat1u) );
  __asm__ __volatile__ ("mfspr %0,0x215"    : "=r" (ibat2l) );
  __asm__ __volatile__ ("mfspr %0,0x214"    : "=r" (ibat2u) );
  __asm__ __volatile__ ("mfspr %0,0x217"    : "=r" (ibat3l) );
  __asm__ __volatile__ ("mfspr %0,0x216"    : "=r" (ibat3u) );
  __asm__ __volatile__ ("mfspr %0,0x219"    : "=r" (dbat0l) );
  __asm__ __volatile__ ("mfspr %0,0x218"    : "=r" (dbat0u) );
  __asm__ __volatile__ ("mfspr %0,0x21B"    : "=r" (dbat1l) );
  __asm__ __volatile__ ("mfspr %0,0x21A"    : "=r" (dbat1u) );
  __asm__ __volatile__ ("mfspr %0,0x21D"    : "=r" (dbat2l) );
  __asm__ __volatile__ ("mfspr %0,0x21C"    : "=r" (dbat2u) );
  __asm__ __volatile__ ("mfspr %0,0x21F"    : "=r" (dbat3l) );
  __asm__ __volatile__ ("mfspr %0,0x21E"    : "=r" (dbat3u) );
  
  fbvideo_write("Machine Status Register value:      ");
  fbvideo_write_hex(msr);
  fbvideo_write(" (");

  if (msr&(MSR_VEC)) {
    fbvideo_write("Altivec Enabled ");
  }
  if (msr&(MSR_POW)) {
    fbvideo_write("Power Management Enabled ");
  }
  if (msr&(MSR_WE)) {
    fbvideo_write("Wait State Enabled ");
  }
  if (msr&(MSR_TGPR)) {
    fbvideo_write("TLB Update in progress ");
  }
  if (msr&(MSR_CE)) {
    fbvideo_write("Critical Interrupt Enabled ");
  }
  if (msr&(MSR_ILE)) {
    fbvideo_write("Little endian mode interrupts ");
  }
  if (msr&(MSR_EE)) {
    fbvideo_write("External interrupts ");
  }
  if (msr&(MSR_PR)) {
    fbvideo_write("Problem State ");
  }
  if (msr&(MSR_FP)) {
    fbvideo_write("Floating Point Enabled ");
  }
  if (msr&(MSR_ME)) {
    fbvideo_write("Machine Check Enabled ");
  }
  if (msr&(MSR_DWE)) {
    fbvideo_write("Debug wait enable "); /* 4xx only and so not used. */
  }
  if (msr&(MSR_BE)) {
    fbvideo_write("Branch Trace ");
  }
  if (msr&(MSR_DE)) {
    fbvideo_write("Debug Exceptions ");
  }
  if (msr&(MSR_FE1)) {
    fbvideo_write("Floating Point Exception Mode 1 ");
  }
  if (msr&(MSR_IP)) {
    fbvideo_write("High Exceptions ");
  } else {
    fbvideo_write("Low Exceptions ");
  }
  if (msr&(MSR_IR)) {
    fbvideo_write("Instruction Relocate ");
  }
  if (msr&(MSR_DR)) {
    fbvideo_write("Data Relocate ");
  }
  if (msr&(MSR_PE)) {
    fbvideo_write("Protection Enabled ");
  }
  if (msr&(MSR_PX)) {
    fbvideo_write("Protection Exclusive Mode ");
  }
  if (msr&(MSR_RI)) {
    fbvideo_write("Recoverable Exception ");
  }
  if (msr&(MSR_LE)) {
    fbvideo_write("Little Endian ");
  }
  fbvideo_write(")\n\n");
  
  fbvideo_write("Processor Version Register value:   ");
  fbvideo_write_hex(pvr);
  fbvideo_write("\n");

  fbvideo_write("Storage Description Register value: ");
  fbvideo_write_hex(sdr1);
  fbvideo_write("\n");
  
  fbvideo_write("IBAT0L: ");
  fbvideo_write_hex(ibat0l);
  fbvideo_write("\r\n");
  fbvideo_write("IBAT0U: ");
  fbvideo_write_hex(ibat0u);
  fbvideo_write("\r\n");
  fbvideo_write("IBAT1L: ");
  fbvideo_write_hex(ibat1l);
  fbvideo_write("\r\n");
  fbvideo_write("IBAT1U: ");
  fbvideo_write_hex(ibat1u);
  fbvideo_write("\r\n");
  fbvideo_write("IBAT2L: ");
  fbvideo_write_hex(ibat2l);
  fbvideo_write("\r\n");
  fbvideo_write("IBAT2U: ");
  fbvideo_write_hex(ibat2u);
  fbvideo_write("\r\n");
  fbvideo_write("IBAT3L: ");
  fbvideo_write_hex(ibat3l);
  fbvideo_write("\r\n");
  fbvideo_write("IBAT3U: ");
  fbvideo_write_hex(ibat3u);
  fbvideo_write("\r\n");
  fbvideo_write("DBAT0L: ");
  fbvideo_write_hex(dbat0l);
  fbvideo_write("\r\n");
  fbvideo_write("DBAT0U: ");
  fbvideo_write_hex(dbat0u);
  fbvideo_write("\r\n");
  fbvideo_write("DBAT1L: ");
  fbvideo_write_hex(dbat1l);
  fbvideo_write("\r\n");
  fbvideo_write("DBAT1U: ");
  fbvideo_write_hex(dbat1u);
  fbvideo_write("\r\n");
  fbvideo_write("DBAT2L: ");
  fbvideo_write_hex(dbat2l);
  fbvideo_write("\r\n");
  fbvideo_write("DBAT2U: ");
  fbvideo_write_hex(dbat2u);
  fbvideo_write("\r\n");
  fbvideo_write("DBAT3L: ");
  fbvideo_write_hex(dbat3l);
  fbvideo_write("\r\n");
  fbvideo_write("DBAT3U: ");
  fbvideo_write_hex(dbat3u);
  fbvideo_write("\r\n");
}
