#ifndef _HAL_PPC_PROCESSOR_H_
#define _HAL_PPC_PROCESSOR_H_
#endif

/* PVR Values
 * The only one we care about is the MPC7455 at present. The others are
 * provided for future extension work.
 */

#define PROCESSOR_VERSION_601           1
#define PROCESSOR_VERSION_603           3
#define PROCESSOR_VERSION_604           4
#define PROCESSOR_VERSION_603e          6
#define PROCESSOR_VERSION_750           8
#define PROCESSOR_VERSION_750FX         0x7000
#define PROCESSOR_VERSION_604e          9
#define PROCESSOR_VERSION_604ev         10
#define PROCESSOR_VERSION_7400          12
#define PROCESSOR_VERSION_7410          0x800C
#define PROCESSOR_VERSION_7450          0x8000
#define PROCESSOR_VERSION_7455          0x8001

/* MSR Values. We do not care about most of these at the moment.
 */

#define MSR_VEC         (1<<25)         /* Enable AltiVec */
#define MSR_POW         (1<<18)         /* Enable Power Management */
#define MSR_WE          (1<<18)         /* Wait State Enable */
#define MSR_TGPR        (1<<17)         /* TLB Update registers in use */
#define MSR_CE          (1<<17)         /* Critical Interrupt Enable */
#define MSR_ILE         (1<<16)         /* Interrupt Little Endian */
#define MSR_EE          (1<<15)         /* External Interrupt Enable */
#define MSR_PR          (1<<14)         /* Problem State / Privilege Level */
#define MSR_FP          (1<<13)         /* Floating Point enable */
#define MSR_ME          (1<<12)         /* Machine Check Enable */
#define MSR_FE0         (1<<11)         /* Floating Exception mode 0 */
#define MSR_SE          (1<<10)         /* Single Step */
#define MSR_DWE         (1<<10)         /* Debug Wait Enable (4xx) */
#define MSR_BE          (1<<9)          /* Branch Trace */
#define MSR_DE          (1<<9)          /* Debug Exception Enable */
#define MSR_FE1         (1<<8)          /* Floating Exception mode 1 */
#define MSR_IP          (1<<6)          /* Exception prefix 0x000/0xFFF */
#define MSR_IR          (1<<5)          /* Instruction Relocate */
#define MSR_DR          (1<<4)          /* Data Relocate */
#define MSR_PE          (1<<3)          /* Protection Enable */
#define MSR_PX          (1<<2)          /* Protection Exclusive Mode */
#define MSR_RI          (1<<1)          /* Recoverable Exception */
#define MSR_LE          (1<<0)          /* Little Endian */

/* Macros to test current MSR status. */

#define MSR_RELOC_OFF	(MSR_ME)

#define MSR_KERNEL      (MSR_ME|MSR_IR|MSR_DR)

#define MSR_PS_TEST(msr) (msr & MSR_PR ? TRUE: FALSE)

/* Processor structures.
 * These include the:  Storage Description Register.
 *                     Block Address Translations.
 */

/* Instruction synchronisation. */
#define sync() \
        __asm__ volatile ("sync")

#define isync() \
        __asm__ volatile ("isync")

/* Storage synchronisation. */
#define eieio() \
        __asm__ volatile ("eieio")


/* Defines for standard required assembly instructions. */

